"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ast_1 = require("javascript-typescript-langserver/lib/ast");
const packages_1 = require("javascript-typescript-langserver/lib/packages");
const symbols_1 = require("javascript-typescript-langserver/lib/symbols");
const typescript_service_1 = require("javascript-typescript-langserver/lib/typescript-service");
const util_1 = require("javascript-typescript-langserver/lib/util");
const opentracing_1 = require("opentracing");
const rxjs_1 = require("rxjs");
const ts = require("typescript");
const lsp_extension_1 = require("@elastic/lsp-extension");
const dependency_manager_1 = require("./dependency-manager");
// import * as rxjs from 'rxjs'
class ExtendedTypescriptService extends typescript_service_1.TypeScriptService {
    // private subscriptions = new rxjs.Subscription()
    constructor(client, options = {}) {
        super(client, options);
        this.client = client;
        this.options = options;
        // @ts-ignore
        // @ts-ignore
        // this.traceModuleResolution = true;
    }
    initialize(params, span) {
        // TODO what about the promise here?
        // TODO run dependencyManager
        return super.initialize(params).map(r => {
            // Must run after super.initialize
            this.dependencyManager = new dependency_manager_1.DependencyManager(this.projectManager, this.packageManager, this.inMemoryFileSystem);
            // Similar to promise then
            // this.subscriptions.add(
            //     Observable.defer(() => {
            if (this.dependencyManager) {
                // this.fileSystem.getWorkspaceFiles().forEach(f => {
                //     if (f.endsWith("package.json")) { // this ensure the file is updated to package manager
                //         this.fileSystem.getTextDocumentContent(f).forEach(c => {
                //             console.log(this.packageManager.packageJsonUris()); // just test code
                //         })
                //     }
                // })
                // fileContentPair.forEach(p => {
                //     this.inMemoryFileSystem.add(p[0], p(1))
                // })
                this.dependencyManager.installDependency();
            }
            else {
                this.logger.error('dependencyManager null');
                // TODO is this the right way?
            }
            return r;
        });
    }
    shutdown(params, span) {
        // this.subscriptions.unsubscribe()
        // TODO shutdown depenency manager
        if (this.dependencyManager) {
            this.dependencyManager.shutdown();
            this.dependencyManager = null;
        }
        else {
            this.logger.error('dependencyManager null');
        }
        if (this.projectManager !== null) {
            return super.shutdown(params);
        }
        else {
            this.logger.error('Server not properly initialized before shutdown');
            return rxjs_1.Observable.of({ op: 'add', path: '', value: null });
        }
    }
    // @ts-ignore
    getHoverForSymbol(info) {
        if (!info) {
            return [];
        }
        // @ts-ignore
        const contents = [];
        // Add declaration without the kind
        const declaration = ts.displayPartsToString(info.displayParts).replace(/^\(.+?\)\s+/, '');
        contents.push({ language: 'typescript', value: this.replaceWorkspaceInString(declaration) });
        if (info.kind) {
            let kind = '**' + info.kind + '**';
            const modifiers = info.kindModifiers
                .split(',')
                // Filter out some quirks like "constructor (exported)"
                .filter(mod => mod &&
                (mod !== ts.ScriptElementKindModifier.exportedModifier ||
                    info.kind !== ts.ScriptElementKind.constructorImplementationElement))
                // Make proper adjectives
                .map(mod => {
                switch (mod) {
                    case ts.ScriptElementKindModifier.ambientModifier:
                        return 'ambient';
                    case ts.ScriptElementKindModifier.exportedModifier:
                        return 'exported';
                    default:
                        return mod;
                }
            });
            if (modifiers.length > 0) {
                kind += ' _(' + modifiers.join(', ') + ')_';
            }
            contents.push(kind);
        }
        // Add documentation
        const documentation = ts.displayPartsToString(info.documentation);
        if (documentation) {
            contents.push(documentation);
        }
        return contents;
    }
    textDocumentFull(params, span = new opentracing_1.Span()) {
        const uri = util_1.normalizeUri(params.textDocument.uri);
        // Ensure files needed to resolve symbols are fetched
        const files = this.projectManager.ensureReferencedFiles(uri, undefined, undefined, span).toArray();
        const symbols = this._getPackageDescriptor(uri).zip(files)
            .mergeMap(res => {
            const fileName = util_1.uri2path(uri);
            const packageDescriptor = res[0];
            const config = this.projectManager.getConfiguration(fileName);
            config.ensureBasicFiles(span);
            const sourceFile = this._getSourceFile(config, fileName, span);
            if (!sourceFile) {
                return [];
            }
            const tree = config.getService().getNavigationTree(fileName);
            return util_1.observableFromIterable(symbols_1.walkNavigationTree(tree))
                .filter(({ tree, parent }) => symbols_1.navigationTreeIsSymbol(tree))
                .map(value => {
                const { tree, parent } = value;
                const symbolInformation = symbols_1.navigationTreeToSymbolInformation(tree, parent, sourceFile, this.root);
                const info = config
                    .getService()
                    .getQuickInfoAtPosition(util_1.uri2path(symbolInformation.location.uri), tree.spans[0].start + 1);
                let contents = '';
                if (info) {
                    contents = this.getHoverForSymbol(info);
                }
                const packageLocator = this.getPackageLocator(packageDescriptor);
                const qname = this.getQnameBySymbolInformation(symbolInformation, packageLocator);
                return {
                    qname,
                    symbolInformation,
                    contents,
                    package: packageLocator
                };
            });
        })
            .toArray();
        let references = rxjs_1.Observable.of([]);
        if (params.reference) {
            references = files
                .mergeMap(() => {
                const fileName = util_1.uri2path(uri);
                // TODO maybe it's better to have a flag
                if (fileName.endsWith('.min.js')) {
                    return [];
                }
                const config = this.projectManager.getConfiguration(fileName);
                config.ensureBasicFiles(span);
                const sourceFile = this._getSourceFile(config, fileName, span);
                if (!sourceFile) {
                    return [];
                }
                return (util_1.observableFromIterable(ast_1.walkMostAST(sourceFile))
                    // Filter Identifier Nodes
                    // Filter defintion self reference
                    // TODO: include string-interpolated references
                    // @ts-ignore
                    .filter((node) => node.kind === ts.SyntaxKind.Identifier)
                    .mergeMap(node => {
                    try {
                        // Find definition for node
                        return rxjs_1.Observable.from(config.getService().getDefinitionAtPosition(sourceFile.fileName, node.pos + 1) || [])
                            .mergeMap(definition => {
                            const symbol = symbols_1.definitionInfoToSymbolDescriptor(definition, this.root);
                            const uri = util_1.path2uri(definition.fileName);
                            const packageDescriptor = this._getPackageDescriptor(uri, span)
                                .defaultIfEmpty(undefined)
                                .map(packageDescriptor => {
                                symbol.package = packageDescriptor;
                                return symbol;
                            });
                            const defintionSourceFile = this._getSourceFile(config, fileName, span);
                            if (!defintionSourceFile) {
                                this.logger.error('Definition Source File not found');
                            }
                            const symbolLoc = this.convertLocation({
                                uri,
                                range: {
                                    start: ts.getLineAndCharacterOfPosition(defintionSourceFile, definition.textSpan.start),
                                    end: ts.getLineAndCharacterOfPosition(defintionSourceFile, ts.textSpanEnd(definition.textSpan)),
                                },
                            });
                            return packageDescriptor.zip(symbolLoc);
                        })
                            .map((pair) => {
                            const symbolDescriptor = pair[0];
                            return {
                                category: lsp_extension_1.ReferenceCategory.UNCATEGORIZED,
                                target: this.getSymbolLocator(symbolDescriptor, pair[1]),
                                location: {
                                    uri: symbols_1.locationUri(sourceFile.fileName),
                                    range: {
                                        start: ts.getLineAndCharacterOfPosition(sourceFile, node.pos),
                                        end: ts.getLineAndCharacterOfPosition(sourceFile, node.end),
                                    },
                                },
                            };
                        });
                    }
                    catch (err) {
                        // Continue with next node on error
                        // Workaround for https://github.com/Microsoft/TypeScript/issues/15219
                        this.logger.error(`textdocument/xreferences: Error getting definition for ${sourceFile.fileName} at offset ${node.pos + 1}`, err);
                        span.log({
                            event: 'error',
                            'error.object': err,
                            message: err.message,
                            stack: err.stack,
                        });
                        return [];
                    }
                }));
            })
                .toArray();
        }
        return symbols
            .zip(references)
            .map(res => {
            const full = { symbols: res[0], references: res[1] };
            return { op: 'add', path: '/-', value: full };
        })
            .startWith({ op: 'add', path: '', value: [] });
    }
    _getHover(params, span = new opentracing_1.Span()) {
        const res = super._getHover(params, span);
        return res.map(h => {
            h.contents = this.replaceWorkspaceInDoc(h.contents);
            return h;
        });
    }
    // Fix go to definition
    _getDefinitionLocations(params, span = new opentracing_1.Span(), goToType = false) {
        const original = super._getDefinitionLocations(params, span, goToType);
        return original.flatMap(location => this.convertLocation(location));
    }
    convertLocation(location) {
        return this.convertUri(location.uri).map(value => {
            location.uri = value;
            return location;
        });
    }
    convertUri(uri) {
        const decodedUri = decodeURIComponent(uri);
        const packageName = packages_1.extractNodeModulesPackageName(uri);
        if (!packageName) {
            return rxjs_1.Observable.of(uri);
        }
        return this._getPackageDescriptor(uri).map(descriptor => {
            const { name, version, repoURL } = descriptor;
            if (!repoURL) {
                return uri;
            }
            let finalURL = 'git://' + repoURL.substring(repoURL.indexOf('://') + 3);
            if (finalURL.endsWith('.git')) {
                finalURL = finalURL.substr(0, finalURL.length - 4);
            }
            let finalVersion = !version ? version : 'master'; // TODO have better syntax
            // TODO use path seperator?
            const moduleString = `node_modules/${name}`;
            let path = decodedUri.substr(decodedUri.indexOf(moduleString) + moduleString.length + 1);
            if (name === 'typescript') {
                finalVersion = 'v' + ts.version;
            }
            else if (uri.startsWith('git://github.com/Microsoft/TypeScript?v')) {
                //  handle the case the the path is already srcgraph's typescript address (see locationUri)
                return uri.replace('#', '/').replace('?', 'blob/');
            }
            else if (name.startsWith('@types/')) {
                // TODO fix version
                finalVersion = 'master';
                path = name.substr(1) + '/' + path;
            }
            else if (finalURL.endsWith('vscode-languageserver-node')) {
                // Following should be our standard package mapping
                const nameMap = {
                    'vscode-languageclient': ['client'],
                    'vscode-languageserver': ['server'],
                    'vscode-languageserver-protocol': ['protocol'],
                    'vscode-languageserver-types': ['types'],
                    'vscode-jsonrpc': ['json-rpc', 'jsonrpc']
                };
                // TODO we might not need this in the future
                finalVersion = encodeURIComponent(`release/${nameMap[name][0]}/${version}`);
                let basePath = nameMap[name][1];
                if (!basePath) {
                    basePath = nameMap[name][0];
                }
                // TODO this is not right because there is no 'lib' dir (it's generated)
                // We'll want symbol:// schema for those whose file are not existed
                path = basePath + '/' + path;
            }
            return `${finalURL}/blob/${finalVersion}/${path}`;
        });
    }
    replaceWorkspaceInDoc(doc) {
        if (doc instanceof Array) {
            for (let i = 0; i < doc.length; i++) {
                // @ts-ignore
                doc[i] = this.replaceWorkspaceInDoc(doc[i]);
            }
        }
        else if (typeof doc === 'string') {
            return this.replaceWorkspaceInString(doc);
        }
        else {
            doc.value = this.replaceWorkspaceInString(doc.value);
        }
        return doc;
    }
    replaceWorkspaceInString(str) {
        let res = str.replace(this.projectManager.getRemoteRoot(), '');
        res = res.replace('"node_modules/', '"'); // TODO consider windows path?
        return res;
    }
    getPackageLocator(packageDescriptor) {
        if (!packageDescriptor) {
            return undefined;
        }
        return {
            name: packageDescriptor.name,
            repoUri: packageDescriptor.repoURL,
            version: packageDescriptor.version
        };
    }
    getSymbolLocator(descriptor, location) {
        return {
            qname: this.getQname(descriptor),
            symbolKind: symbols_1.stringtoSymbolKind(descriptor.kind),
            path: descriptor.filePath,
            package: this.getPackageLocator(descriptor.package),
            location,
        };
    }
    textDocumentEdefinition(params, span = new opentracing_1.Span()) {
        return this._getSymbolLocationInformations(params, span)
            .flatMap(symbol => this.getSymbolLocatorFromLocationInformation(symbol))
            .map(symbol => ({ op: 'add', path: '/-', value: symbol }))
            .startWith({ op: 'add', path: '', value: [] });
    }
    getSymbolLocatorFromLocationInformation(locationInfo) {
        // let location: Observable<Location> | undefined
        if (locationInfo.location) {
            return this.convertLocation(locationInfo.location).map(l => ({
                qname: this.getQname(locationInfo.symbol),
                symbolKind: symbols_1.stringtoSymbolKind(locationInfo.symbol.kind),
                path: locationInfo.symbol.filePath,
                package: this.getPackageLocator(locationInfo.symbol.package),
                location: l
            }));
        }
        return rxjs_1.Observable.of({
            qname: this.getQname(locationInfo.symbol),
            symbolKind: symbols_1.stringtoSymbolKind(locationInfo.symbol.kind),
            path: locationInfo.symbol.filePath,
            package: this.getPackageLocator(locationInfo.symbol.package),
            location: undefined
        });
    }
    getQname(desc) {
        let prefix = '';
        // if (desc.package) {
        //     prefix += desc.package.name + '.'
        // } else {
        //     prefix = 'unknown.'
        // }
        // if (desc.name === "Error") {
        //     console.log("")
        // }
        //  TODO check with type
        const fileName = this.getFileName(desc.filePath);
        // const simpleName = this.getSimpleFileName(fileName)
        // if (desc.filePath !== '') {
        //     prefix += simpleName + '.'
        // }
        if (desc.containerName !== '' && desc.containerName.indexOf(fileName) === -1) {
            prefix += this.cleanContainerName(desc.containerName) + '.';
        }
        return prefix + desc.name;
    }
    getQnameBySymbolInformation(info, packageLocator) {
        let prefix = '';
        // if (packageLocator && packageLocator.name && packageLocator.name !== '') {
        //     prefix += packageLocator.name + '.'
        // } else {
        //     prefix = 'unknown'
        // }
        const fileName = this.getFileName(info.location.uri);
        // const simpleName = this.getSimpleFileName(fileName)
        // if (info.location.uri !== '') {
        //     prefix += simpleName + '.'
        // }
        if (info.containerName && info.containerName !== '' && info.containerName.indexOf(fileName) === -1) {
            prefix += this.cleanContainerName(info.containerName) + '.';
        }
        return prefix + info.name;
    }
    getFileName(pathOrUri) {
        // @ts-ignore
        return pathOrUri.split('\\').pop().split('/').pop();
        // return fileName.substr(0, fileName.indexOf('.'))
    }
    cleanContainerName(name) {
        return name.split('"').join('').split('\\').join('.').split('/').join('.');
    }
}
exports.ExtendedTypescriptService = ExtendedTypescriptService;
//# sourceMappingURL=extended-typescript-service.js.map